package net.shadowfacts.cacao

import net.shadowfacts.cacao.geometry.Point
import net.shadowfacts.cacao.util.MouseButton
import net.shadowfacts.cacao.view.View
import no.birkett.kiwi.Solver
import java.util.*

/**
 * A Window is the object at the top of a Cacao view hierarchy. It occupies the entirety of the Minecraft screen size
 * and provides the base coordinate system for its view hierarchy.
 *
 * The Window owns the Kiwi [Solver] object used for layout by all of its views.
 *
 * @author shadowfacts
 */
class Window {

	/**
	 * The screen that this window belongs to.
	 * Not initialized until this window is added to a screen, using it before that point will throw a runtime exception.
	 */
	lateinit var screen: CacaoScreen

	var solver = Solver()

	// _views is the internal, mutable object, since we only want it to be mutated by the add/removeView methods
	private val _views = LinkedList<View>()
	/**
	 * The list of top-level views in this window.
	 * This list should never be modified directly, only by calling the [addView]/[removeView] methods.
	 */
	val views: List<View> = _views

	private var viewsSortedByZIndex: List<View> = listOf()

	/**
	 * Convenience method that removes this window from its [screen].
	 */
	fun removeFromScreen() {
		screen.removeWindow(this)
	}

	/**
	 * Adds the given view as a top-level view in this window.
	 *
	 * @param view The view to add.
	 * @return The same view, as a convenience.
	 */
	fun <T: View> addView(view: T): T {
		_views.add(view)
		viewsSortedByZIndex = views.sortedBy(View::zIndex)

		view.window = this
		view.solver = solver

		view.wasAdded()

		return view
	}

	/**
	 * Finds all views in this window at the given point.
	 *
	 * @param point The point to find views at, in the coordinate system of the window.
	 * @return All views that contain the point.
	 */
	fun viewsAtPoint(point: Point): List<View> {
		return views.filter { point in it.frame }
	}

	/**
	 * Attempts to find a top level view of this window that contains the given point.
	 * If there are multiple overlapping views, which one this method returns is undefined.
	 * [viewsAtPoint] may be used, and the resulting List sorted by [View.zIndex].
	 *
	 * @param point The point to find views at, in the coordinate system of the window.
	 * @return the Veiw, if any, that contain the given point.
	 */
	fun viewAtPoint(point: Point): View? {
		return views.firstOrNull { point in it.frame }
	}

	/**
	 * Instructs the solver to solve all of the provided constraints.
	 * Should be called after the view hierarchy is setup.
	 */
	fun layout() {
		solver.updateVariables()
		views.forEach(View::didLayout)
	}

	/**
	 * Draws this window and all of its views.
	 * This method is called by [CacaoScreen] and generally shouldn't be called directly.
	 *
	 * @param mouse The point in the coordinate system of the window.
	 * @param delta The time elapsed since the last frame.
	 */
	fun draw(mouse: Point, delta: Float) {
		viewsSortedByZIndex.forEach {
			val mouseInView = Point(mouse.x - it.frame.left, mouse.y - it.frame.top)
			it.draw(mouseInView, delta)
		}
	}

	/**
	 * Called when a mouse button is clicked and this is the active window.
	 * This method is called by [CacaoScreen] and generally shouldn't be called directly.
	 *
	 * @param point The point in the window of the click.
	 * @param mouseButton The mouse button that was used to click.
	 * @return Whether the mouse click was handled by a view.
	 */
	fun mouseClicked(point: Point, mouseButton: MouseButton): Boolean {
		val view = viewsAtPoint(point).maxBy(View::zIndex)
		if (view != null) {
			val pointInView = Point(point.x - view.frame.left, point.y - view.frame.top)
			return view.mouseClicked(pointInView, mouseButton)
		}
		return false
	}

}